# Generated by Django 4.0 on 2021-12-15 06:20

import django.contrib.postgres.fields
import django.contrib.postgres.indexes
import django.contrib.postgres.search
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.db.models.manager
import urlid_graph.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Entity',
            fields=[
                ('uuid', models.UUIDField(primary_key=True, serialize=False)),
                ('base_url', models.TextField()),
                ('name', models.TextField()),
                ('version', models.TextField()),
            ],
            managers=[
                ('brasilio', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='ObjectRepository',
            fields=[
                ('uuid', models.UUIDField(primary_key=True, serialize=False)),
                ('search_data', django.contrib.postgres.search.SearchVectorField(null=True)),
                ('entity', models.ForeignKey(db_constraint=False, db_index=False, on_delete=django.db.models.deletion.DO_NOTHING, to='urlid_graph.entity')),
            ],
            bases=(urlid_graph.models.ObjectMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ElementConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('config_type', models.PositiveSmallIntegerField(choices=[(1, 'entity'), (2, 'relationship'), (3, 'property'), (4, 'stats')])),
                ('name', models.TextField()),
                ('parent_type', models.PositiveSmallIntegerField(choices=[(1, 'entity'), (2, 'relationship')], null=True)),
                ('parent_name', models.TextField(null=True)),
                ('label', models.TextField()),
                ('data', models.JSONField()),
            ],
            options={
                'unique_together': {('config_type', 'name', 'parent_type', 'parent_name')},
            },
        ),
        migrations.CreateModel(
            name='EntityConfig',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('urlid_graph.elementconfig',),
        ),
        migrations.CreateModel(
            name='EntityPropertyConfig',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('urlid_graph.elementconfig',),
        ),
        migrations.CreateModel(
            name='RelationshipConfig',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('urlid_graph.elementconfig',),
        ),
        migrations.CreateModel(
            name='RelPropertyConfig',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('urlid_graph.elementconfig',),
        ),
        migrations.CreateModel(
            name='StatsConfig',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('urlid_graph.elementconfig',),
        ),
        migrations.CreateModel(
            name='SavedGraph',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('edges', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), size=None)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('user', 'name')},
            },
        ),
        migrations.AddIndex(
            model_name='objectrepository',
            index=django.contrib.postgres.indexes.GinIndex(fields=['search_data'], name='urlid_graph_search__f39787_gin'),
        ),
        migrations.AlterUniqueTogether(
            name='objectrepository',
            unique_together={('entity', 'uuid')},
        ),
    ]



# Generated by Django 4.0.5 on 2022-07-31 22:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='elementconfig',
            options={'verbose_name': 'Element configuration', 'verbose_name_plural': 'Element configurations'},
        ),
        migrations.AlterModelOptions(
            name='entity',
            options={'verbose_name': 'Entity', 'verbose_name_plural': 'Entities'},
        ),
        migrations.AlterModelOptions(
            name='objectrepository',
            options={'verbose_name': 'Object repository', 'verbose_name_plural': 'Object repositories'},
        ),
        migrations.AlterModelOptions(
            name='savedgraph',
            options={'verbose_name': 'Saved graph', 'verbose_name_plural': 'Saved graphs'},
        ),
        migrations.AlterModelManagers(
            name='entity',
            managers=[
            ],
        ),
        migrations.AlterUniqueTogether(
            name='objectrepository',
            unique_together=set(),
        ),
    ]



from collections import OrderedDict
from copy import deepcopy
from functools import lru_cache

from django.core.exceptions import ObjectDoesNotExist
from django.urls import reverse
from rest_framework import serializers
from rest_framework.validators import UniqueTogetherValidator

from urlid_graph import settings as urlid_graph_settings

from .exceptions import InvalidEdgeIdException
from .formatting import format_property_value
from .graph_db import parse_edge_id
from .models import ObjectRepository, RelationshipConfig, RelPropertyConfig, SavedGraph


def _sorted_properties(props):
    return OrderedDict([(k, v) for k, v in sorted(props.items(), key=lambda p: p[0])])


class RelationshipSerializer:

    name = None  # Must be set

    def __init__(self, obj):
        self.obj = obj
        self._data = None

    @property
    def label(self):
        return self.name

    @property
    def data(self):
        if self._data is None:
            edge_label = self.obj["label"]
            props = {}
            for name, value in self.obj["properties"].items():
                key_config = RelPropertyConfig.objects.filter(parent_name=edge_label, name=name).first()
                props[key_config.label if key_config is not None else name] = value
            self._data = _sorted_properties(props)
        return self._data


def subclasses(cls):
    """Return all subclasses of a class, recursively"""
    children = cls.__subclasses__()
    return set(children).union(
        set(grandchild for child in children for grandchild in subclasses(child))
    )


@lru_cache(maxsize=1024)
def get_relationship_serializer(name):
    classes = {
        Class.name: Class for Class in subclasses(RelationshipSerializer)
        if Class != RelationshipSerializer
    }
    return classes.get(name, RelationshipSerializer)


class ConfigSerializer(serializers.Serializer):
    options = serializers.SerializerMethodField()
    endpoints = serializers.SerializerMethodField()

    def get_options(self, obj):
        from urlid_graph.network_vis_config import get_entity_node_config, graph_vis_options

        options = deepcopy(graph_vis_options)

        for entity in obj["entities"]:
            options["groups"].update(get_entity_node_config(entity))

        return options

    def get_endpoints(self, obj):
        return {"search": reverse("graph_api:search")}


class UUIDListSerializer(serializers.Serializer):
    uuids = serializers.ListField(child=serializers.UUIDField())

    def validate_uuids(self, value):
        if len(value) > urlid_graph_settings.NODES_CSV_SIZE_LIMIT:
            raise serializers.ValidationError(
                f"Number of nodes exceeded maximum {urlid_graph_settings.NODES_CSV_SIZE_LIMIT}"
            )
        return value


class EdgeIdField(serializers.Field):
    edge_id = serializers.CharField()

    def to_representation(self, value):
        return value

    def to_internal_value(self, value):
        try:
            edge_info = parse_edge_id(value)
        except InvalidEdgeIdException:
            raise serializers.ValidationError("Edge(s) not in correct format: 'label|from_id|to_id'")
        serializers.UUIDField().to_internal_value(edge_info.from_uuid)
        serializers.UUIDField().to_internal_value(edge_info.to_uuid)

        return value


class SavedGraphSerializer(serializers.ModelSerializer):
    user = serializers.HiddenField(default=serializers.CurrentUserDefault())
    edges = serializers.ListField(child=EdgeIdField(), allow_empty=False)
    name = serializers.CharField()
    created_at = serializers.DateTimeField(
        required=False,
        format="%d/%m/%Y %H:%M",
        read_only=True,
    )

    def to_representation(self, obj):
        response = super().to_representation(obj)
        if self.context["request"].method == "GET":
            response.pop("edges")

        return response

    def get_name(self, value):
        return value.strip()

    def create(self, validated_data):
        return SavedGraph.objects.create(**validated_data)

    class Meta:
        model = SavedGraph
        fields = ["pk", "user", "name", "edges", "created_at"]
        validators = [UniqueTogetherValidator(queryset=SavedGraph.objects.all(), fields=["user", "name"])]


class EdgeSerializer(serializers.Serializer):
    name = serializers.CharField(source="label")
    to = serializers.CharField()
    id = serializers.CharField()
    label = serializers.SerializerMethodField()

    def __init__(self, *args, **kwargs):
        # from is a reserved word and can't be used to declare fields the same
        # way as the others are declared
        self.fields["from"] = serializers.CharField()
        super().__init__(*args, **kwargs)

    def get_label(self, obj):
        edge_label = obj["label"]
        Serializer = get_relationship_serializer(edge_label)
        label = Serializer(obj).label
        if label != edge_label:
            return label

        try:
            return RelationshipConfig.objects.get_by_name(edge_label).label
        except ObjectDoesNotExist:
            return edge_label


class DetailedEdgeSerializer(EdgeSerializer):
    properties = serializers.SerializerMethodField()

    def get_properties(self, obj):
        Serializer = get_relationship_serializer(obj["label"])
        return Serializer(obj).data


class NodeSerializer(serializers.ModelSerializer):
    group = serializers.CharField(source="entity.name")
    label = serializers.CharField()
    id = serializers.CharField(source="uuid")

    class Meta:
        model = ObjectRepository
        fields = ["id", "group", "label"]

    def to_representation(self, instance):
        result = super().to_representation(instance)
        result["properties"] = instance.properties
        return result


class DetailedNodeSerializer(NodeSerializer):
    properties = serializers.SerializerMethodField()

    class Meta:
        model = NodeSerializer.Meta.model
        fields = NodeSerializer.Meta.fields + ["properties"]

    def get_properties(self, obj):
        # TODO: unify with Model.serialize()
        props = {}
        for name, value in obj.properties.items():
            key = obj.get_label_for_property(name)
            props[key] = format_property_value(name, key, value)
        return _sorted_properties(props)


class FullPropertiesNodeSerializer(DetailedNodeSerializer):
    full_properties = serializers.SerializerMethodField()

    class Meta:
        model = DetailedNodeSerializer.Meta.model
        fields = DetailedNodeSerializer.Meta.fields + ["full_properties"]

    def get_full_properties(self, obj):
        # TODO: unify with Model.serialize()
        props = {}
        for name, value in obj.full_properties.items():
            key = obj.get_label_for_property(name)
            props[key] = []
            for prop_value in value:
                prop_value["value"] = format_property_value(name, key, prop_value["value"])
                props[key].append(prop_value)
        return props


class RelationshipArgumentsSerializer(serializers.Serializer):
    depth = serializers.IntegerField(default=0, min_value=0, max_value=urlid_graph_settings.RELATIONSHIP_DEPTH_LIMIT)
    inbound = serializers.BooleanField(default=True)
    outbound = serializers.BooleanField(default=True)

    def validate(self, attrs):
        if attrs["inbound"] is False and attrs["outbound"] is False:
            raise serializers.ValidationError("Parameters inbound and outbound can not be both False")

        return attrs


class AllNodesRelationshipsSerializer(serializers.Serializer):
    uuids = serializers.ListField(
        child=serializers.UUIDField(), max_length=urlid_graph_settings.ALL_NODES_RELATIONSHIPS_CHUNK_SIZE, min_length=1
    )

    def validate(sel, attrs):
        return {"uuids": tuple(str(u) for u in attrs["uuids"])}



from django.urls import path, re_path

from . import views

list_actions = {"get": "list", "post": "create"}
single_actions = {
    "get": "retrieve",
    "put": "update",
    "patch": "partial_update",
    "delete": "destroy",
}


app_name = "graph_api"

urlpatterns = [
    path("config", views.GraphDataVisConfig.as_view(), name="config"),
    path(
        "dataset/<str:slug>/<uuid:object_uuid>",
        views.DatasetFilter.as_view(),
        name="dataset-filter",
    ),
    path(
        "edge/<str:edge_id>",
        views.GraphEdgeDetailEndpoint.as_view(),
        name="edge_detail",
    ),
    path(
        "export-properties-csv",
        views.ExportVerticesCSVView.as_view(),
        name="export-properties-csv",
    ),
    path(
        "node/<uuid:uuid>",
        views.GraphNodeDetailEndpoint.as_view(),
        name="node_detail",
    ),
    path(
        "node/<uuid:uuid>/relationships",
        views.NodeRelationshipsEndpoint.as_view(),
        name="node_rels",
    ),
    path(
        "node/relationships",
        views.AllNodesRelationshipsEndpoint.as_view(),
        name="all_nodes_rels",
    ),
    path(
        "save-graph",
        views.SavedGraphViewSet.as_view(list_actions),
        name="save-graph",
    ),
    re_path(
        "^save-graph/(?P<pk>\d*)$",
        views.SavedGraphViewSet.as_view(single_actions),
        name="save-graph",
    ),  # noqa
    re_path(
        "^saved-graph-detail/(?P<pk>\d*)$",
        views.SavedGraphDetails.as_view(),
        name="saved-graph-detail",
    ),  # noqa
    path("search", views.SearchOnGraphEndpoint.as_view(), name="search"),
    path(
        "shortest-path/<uuid:from_uuid>/<uuid:to_uuid>",
        views.ShortestPathEndpoint.as_view(),
        name="shortest_path",
    ),
]



import json

import rows
from django.core.management.base import BaseCommand

from urlid_graph.models import ElementConfig


class Command(BaseCommand):
    help = "Export elements' configurations"

    def add_arguments(self, parser):
        parser.add_argument("output_filename")

    def handle(self, *args, **options):
        writer = rows.utils.CsvLazyDictWriter(options["output_filename"])
        for element in ElementConfig.objects.all().iterator():
            writer.writerow(
                {
                    key: value if key != "data" else json.dumps(value)
                    for key, value in element.__dict__.items()
                    if key != "_state"
                }
            )
        writer.close()



import re
from datetime import date, datetime

from django.conf import settings
from django.contrib.humanize.templatetags.humanize import intcomma
from django.template.defaultfilters import date as django_date

REGEXP_NOT_NUMBERS = re.compile("[^0-9*]")


def convert_date(value):
    if isinstance(value, str):
        return value
    return datetime.strptime(value, "%Y-%m-%d").date()


def convert_datetime(value):
    if isinstance(value, str):
        return value
    return datetime.strptime(value, "%Y-%m-%d %H:%M:%S")


def convert_bool(value):
    return bool(int(value))


def format_cnpj(value):
    value = (value or "").strip()
    if not value:
        return None
    value = f"{int(REGEXP_NOT_NUMBERS.sub('', value)):014d}"

    return f"{value[:2]}.{value[2:5]}.{value[5:8]}/{value[8:12]}-{value[12:14]}"


def format_cpf(value):
    value = (value or "").strip()
    if not value:
        return None
    value = REGEXP_NOT_NUMBERS.sub("", value)
    if "*" not in value:
        value = f"{int(value):011d}"
    elif len(value) < 11:
        value = "0" * (11 - len(value)) + value

    return f"{value[:3]}.{value[3:6]}.{value[6:9]}-{value[9:11]}"


def format_titulo_eleitoral(value):
    value = (value or "").strip()
    if not value:
        return None
    value = f"{int(REGEXP_NOT_NUMBERS.sub('', value)):012d}"

    return f"{value[:4]}.{value[4:8]}.{value[8:10]}-{value[10:12]}"


def format_property_value(name, label, value):
    if isinstance(value, datetime):
        return django_date(value, settings.DATETIME_FORMAT)
    elif isinstance(value, date):
        return django_date(value, settings.DATE_FORMAT)
    elif name == "cnpj":
        return format_cnpj(value)
    elif name.startswith("cpf"):
        return format_cpf(value)
    elif name == "titulo_eleitoral":
        return format_titulo_eleitoral(value)
    elif name == "numero_sequencial":
        return intcomma(value)
    else:
        return value



import traceback

from django.apps import apps
from django.core.management.base import BaseCommand

from urlid_graph.log import IndexObjectsLogger
from urlid_graph.models import ObjectModelMixin, ObjectRepository


class Command(BaseCommand):
    help = "Index objects from a specific model"

    def add_arguments(self, parser):
        parser.add_argument("--start-id", type=int, default=1)
        parser.add_argument("--batch-size", type=int, default=100_000)
        parser.add_argument("app_name")
        parser.add_argument("model")

    def handle(self, *args, **options):
        app_name = options["app_name"]
        model = options["model"]
        Model = apps.get_model(app_name, model)
        if not issubclass(Model, ObjectModelMixin):
            raise ValueError("Model '{}' doesn't inherit from `ObjectModelMixin`".format(model))

        self.logger = IndexObjectsLogger(description=f"{app_name}.{model}")
        print(f"Starting job {self.logger.job.id}")

        ok = True
        try:
            self.logger.start_step("delete-objects")
            result = ObjectRepository.objects.filter(entity_id=Model._meta.entity_uuid).delete()
            self.logger.finish_step("delete-objects", message=f"{result[0]} objects deleted")

            self.logger.start_step("index")
            ObjectRepository.objects.index(
                Model,
                start_id=options["start_id"],
                batch_size=options["batch_size"],
                callback=self.logger.progress,
            )
            self.logger.finish_step("index")

        except:  # noqa
            self.logger.error(traceback.format_exc())
            ok = False

        else:
            self.logger.finish()

        return str(ok)  # Used by import_data when calling this command programatically



import textwrap
import traceback
from collections import OrderedDict
from pathlib import Path

import rows
from django.core.management.base import BaseCommand
from django.db import connections
from psycopg2.sql import SQL, Identifier

import urlid_graph.settings as urlid_graph_settings
from urlid_graph.log import ImportRelationshipsLogger
from urlid_graph.utils import DatabaseConnection, random_name, read_total_size


class Command(BaseCommand):
    help = "Import relations to graph database"

    def add_arguments(self, parser):
        parser.add_argument("--batch-size", type=int, default=10000)
        parser.add_argument("--chunk-size", type=int, default=8388608)
        parser.add_argument("--debug", action="store_true")
        parser.add_argument("--disable-autovacuum", action="store_true")
        parser.add_argument("--no-drop-table", action="store_true")
        parser.add_argument("--sample-size", type=int, default=3000)
        parser.add_argument("relationship")
        parser.add_argument("input_filename")

    def handle(self, *args, **options):
        debug = options.get("debug", False)
        disable_autovacuum = options["disable_autovacuum"]
        relationship = options.pop("relationship")
        input_filename = options.pop("input_filename")

        django_connection = connections[urlid_graph_settings.GRAPH_DATABASE]
        django_connection.connect()
        self.db = DatabaseConnection(connection=django_connection.connection, debug=debug)
        # TODO: `.lower()` is needed so rows' slug won't mess with the name
        temp_table_name = "tmp_" + random_name().lower()

        self.logger = ImportRelationshipsLogger(description=f"{Path(input_filename).name} -> {relationship}")
        print(f"Starting job {self.logger.job.id}")

        ok = True
        try:
            with self.logger.step("pre-import", message_start="Preparing table to import"):
                self.logger.message("Disabling sync commit")
                self.db.disable_sync_commit()

                if disable_autovacuum:
                    self.logger.message(f"Disabling autovacuum for {table_name}")
                    self.db.disable_autovacuum()

            try:
                self.execute_import_data(relationship, input_filename, temp_table_name, *args, **options)
            except:  # noqa
                self.logger.error(traceback.format_exc())
                ok = False

            for relname in (
                f"{urlid_graph_settings.GRAPH_NAME}.ag_vertex",
                f"{urlid_graph_settings.GRAPH_NAME}.ag_edge",
            ):
                self.logger.message(f"Running VACUUM ANALYZE on {relname}")
                self.db.vacuum_analyze(relname)

        except:  # noqa
            self.logger.error(traceback.format_exc())
            ok = False

        finally:
            with self.logger.step("post-import"):
                if disable_autovacuum:
                    self.logger.message(f"Enabling autovacuum for {table_name}")
                    self.db.enable_autovacuum()
                self.logger.message("Enabling sync commit")
                self.db.enable_sync_commit()

        self.logger.finish()
        return str(ok)  # Used by import_data when calling this command programatically

    def create_properties(self, header, type_):
        # TODO: use `rows.properties.{field}` instead
        props = []
        for field in header:
            props.append(f"ON {type_.upper()} SET r.{field} = row.{field}")

        return "\r\n".join(props)

    def create_merge_query(self, schema, relation_name, temp_table_name):
        header = [field_name for field_name in schema.keys() if field_name != "id" and "node_uuid" not in field_name]
        props_create = self.create_properties(header, type_="create")
        props_match = self.create_properties(header, type_="match")
        # https://neo4j.com/developer/kb/understanding-how-merge-works/

        # TODO: may create a VLABEL for each entity and import the objects
        # using this VLABEL (not `object`) - it may not be possible depending
        # on the extraction file format.
        return textwrap.dedent(
            """
            LOAD FROM {temp_table_name} as row
            WITH row WHERE $1 < row.id AND row.id <= $2
            MERGE (obj1:object {{uuid: row.from_node_uuid}})
            MERGE (obj2:object {{uuid: row.to_node_uuid}})
            MERGE (obj1)-[r:{relation}]->(obj2)
            {create}
            {match}
            """.format(
                relation=relation_name,
                temp_table_name=temp_table_name,
                create=props_create,
                match=props_match,
            )
        )

    def create_table(self, schema, table_name):
        # Create table manually so we can inject an `id` column to control the
        # batch import data.

        schema = schema.copy()  # Do not change original object
        schema["id"] = rows.fields.IntegerField
        sql = rows.utils.pg_create_table_sql(schema, table_name=table_name, unlogged=True)
        # TODO: change the way we change this behavior (serial)
        sql = sql.replace('"id" BIGINT', '"id" SERIAL').replace("id BIGINT", "id SERIAL")
        self.db.execute_query(sql)

    def optimize_data_table(self, table_name):
        self.db.execute_queries(
            [
                f"ALTER TABLE {table_name} ADD PRIMARY KEY (id)",
                f"VACUUM ANALYZE {table_name}",
            ],
        )

    def ensure_graph_requisites(self, relation_name):
        self.db.execute_query(
            SQL("CREATE ELABEL IF NOT EXISTS {}").format(Identifier(relation_name)),
        )
        self.db.execute_query(
            SQL("CREATE VLABEL IF NOT EXISTS {}").format(Identifier("object")),
        )
        self.db.execute_query(
            SQL("CREATE PROPERTY INDEX IF NOT EXISTS idx_vlabel_object_uuid ON object (uuid)"),
        )

    def import_relations(self, schema, relation_name, temp_table_name, n_rows, batch_size):
        cypher_statement = self.create_merge_query(schema, relation_name, temp_table_name)
        statement_name = f"urlid_tmp_rel_{random_name(10)}"
        self.db.execute_query(f"PREPARE {statement_name}(int, int) AS {cypher_statement}")
        for min_id in range(0, n_rows + 1, batch_size):
            self.db.execute_query(f"EXECUTE {statement_name} ({min_id}, {min_id + batch_size})")
            self.logger.progress(done=min(min_id + batch_size, n_rows), total=n_rows)
        self.db.execute_query(f"DEALLOCATE {statement_name}")

    def execute_import_data(self, relation_name, input_filename, temp_table_name, *args, **options):
        self.logger.start_step("import-table-data")
        self.logger.message(f"Creating table {temp_table_name}")
        schema = OrderedDict(
            [
                ("from_node_uuid", rows.fields.UUIDField),
                ("to_node_uuid", rows.fields.UUIDField),
                ("relationship", rows.fields.TextField),
                ("properties", rows.fields.JSONField),
            ]
        )
        self.create_table(schema, temp_table_name)
        file_size = read_total_size(input_filename)

        def update_progress(done, total):
            # done is actually the last step done and total the total_done
            self.logger.progress(done=total, total=file_size)

        self.logger.message("Importing data")
        rows_ouput = rows.utils.pgimport(
            input_filename,
            encoding="utf-8",
            dialect="excel",
            table_name=temp_table_name,
            create_table=False,
            database_uri=urlid_graph_settings.GRAPH_DATABASE_URL,
            schema=schema,
            callback=update_progress,
            chunk_size=options["chunk_size"],
        )
        total_imported, batch_size = rows_ouput["rows_imported"], options["batch_size"]
        self.logger.message(f"Imported {total_imported} rows to '{temp_table_name}'")

        self.logger.message("Optimizing table for reading")
        self.optimize_data_table(temp_table_name)
        self.logger.finish_step("import-table-data")

        self.logger.start_step("import-relationships")
        self.logger.message(f"Ensuring elabel {relation_name} exists")
        self.ensure_graph_requisites(relation_name)

        self.import_relations(schema, relation_name, temp_table_name, total_imported, batch_size)

        if not options["no_drop_table"]:
            self.logger.message(f"Deleting table '{temp_table_name}'")
            self.db.delete_table(temp_table_name)
        self.logger.finish_step("import-relationships")



import rows
from django.core.management.base import BaseCommand

from urlid_graph.models import ElementConfig


class Command(BaseCommand):
    help = "Import entities, properties and relationships config"

    def add_arguments(self, parser):
        parser.add_argument("input_filename")

    def handle(self, *args, **options):
        config_rows = rows.import_from_csv(
            options["input_filename"],
            force_types={
                "config_type": rows.fields.IntegerField,
                "data": rows.fields.JSONField,
                "label": rows.fields.TextField,
                "name": rows.fields.TextField,
                "parent_name": rows.fields.TextField,
                "parent_type": rows.fields.IntegerField,
            },
        )

        for row in config_rows:
            cfg, created = ElementConfig.objects.update_or_create(
                config_type=row.config_type,
                name=row.name,
                parent_type=row.parent_type or None,
                parent_name=row.parent_name,
                defaults={"label": row.label, "data": row.data},
            )

            print(
                f"{'CREATED' if created else 'UPDATED'} cfg {cfg.label} para {cfg.get_config_type_display()} {cfg.name}"
            )



# Generated by Django 4.0.6 on 2022-08-09 10:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0005_alter_joblog_options_alter_logstep_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='elementconfig',
            name='config_type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'entity'), (2, 'relationship'), (3, 'property'), (4, 'stats'), (5, 'custom')]),
        ),
    ]



import datetime
import re
import uuid
from collections import defaultdict
from functools import lru_cache
from textwrap import dedent
from urllib.parse import urljoin

import django.db.models.options as options
from cached_property import cached_property
from django.conf import settings
from django.core.cache import cache
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchQuery, SearchRank, SearchVectorField
from django.core.exceptions import ObjectDoesNotExist
from django.db import connections, models
from django.template import Context, Template

from urlid_graph import settings as urlid_graph_settings
from urlid_graph.utils import random_name

options.DEFAULT_NAMES = options.DEFAULT_NAMES + ("entity_name", "entity_uuid", "search_fields", "slug")


REGEXP_CAMEL_CASE = re.compile("([A-Z])")


def get_urlid_database_uri():
    db = settings.DATABASES[urlid_graph_settings.DJANGO_DATABASE]
    return f"postgres://{db['USER']}:{db['PASSWORD']}@{db['HOST']}:{db['PORT']}/{db['NAME']}"


def slug_from_model_name(name):
    """
    >>> slug_from_model_name('SomeDataset')
    'some-dataset'
    >>> slug_from_model_name('MyTestModel')
    'my-test'
    """

    name = REGEXP_CAMEL_CASE.sub(r"-\1", name).lower()
    if name[0] == "-":
        name = name[1:]
    if name[-1] == "-":
        name = name[:-1]
    if name.endswith("-model"):
        name = name[:-6]
    return name


@lru_cache(maxsize=64)
def get_template(template_code):
    return Template(template_code)


class EntityQuerySet(models.QuerySet):
    def by_name(self, names):
        return self.filter(name__in=[name.lower().strip() for name in names])


class Entity(models.Model):
    objects = EntityQuerySet.as_manager()

    uuid = models.UUIDField(primary_key=True)
    base_url = models.TextField(blank=False, null=False)
    name = models.TextField(blank=False, null=False)
    version = models.TextField(blank=False, null=False)

    class Meta:
        verbose_name = "Entity"
        verbose_name_plural = "Entities"

    @property
    def url(self):
        return urljoin(self.base_url, f"/{self.name}/v{self.version}/")

    @property
    def label_properties(self):
        return getattr(self.config, "label_properties", None) or []

    @property
    def label_template(self):
        return getattr(self.config, "label_template", None) or None

    @property
    def graph_node_conf(self):
        return getattr(self.config, "graph_node_conf", None) or {}

    @property
    def config(self):
        try:
            return EntityConfig.objects.for_entity(self.name)
        except ObjectDoesNotExist:
            return None

    @property
    def label(self):
        return getattr(self.config, "label", None) or self.name

    def __str__(self):
        return f"Entity {self.url}"

    def get_model(self):
        for Model in ObjectModelMixin.__subclasses__():
            if str(Model._meta.entity_uuid) == str(self.uuid):
                return Model


class ObjectMixin:
    """Mixin used by Object and ObjectRepository classes"""

    @property
    def url(self):
        # TODO: change
        return urljoin(self.entity.url, f"{self.uuid}/")

    def _get_objects(self):
        ObjectModel = self.entity.get_model()
        if ObjectModel is None:
            raise ValueError("Model class not found fo entity UUID {}".format(self.entity.uuid))

        return ObjectModel.objects.filter(object_uuid=self.uuid).order_by(models.F("updated_at").asc(nulls_first=True))

    @cached_property
    def properties(self):
        data = {}
        for obj in self._get_objects():
            data.update(obj.serialize())
        return data

    @cached_property
    def raw_properties(self):
        data = {}
        for obj in self._get_objects():
            data.update(obj.raw_serialize())
        return data

    @cached_property
    def full_properties(self):
        data = defaultdict(list)
        for obj in self._get_objects():
            for key, value in obj.serialize():
                data[key].append(
                    {
                        "value": value,
                        "value_type": 1,  # TODO: change?
                        "source": "?",  # TODO: change?
                        "value_datetime": obj.updated_at.isoformat() if obj.updated_at else None,
                    }
                )
        return data

    @property
    def label(self):
        label_template = self.entity.label_template
        props = self.raw_properties
        if label_template is not None:
            return get_template(label_template).render(Context(props))
        else:
            for prop_name in self.entity.label_properties:
                label = props.get(prop_name)
                if label:
                    return label
        return f"{self.entity.name}: {self.uuid}"

    def get_label_for_property(self, prop_name):
        key = f"urlid_graph:ElementConfig:{self.entity.name}:{prop_name}"
        cached = cache.get(key)
        if cached is not None:
            return cached
        else:
            try:
                config = EntityPropertyConfig.objects.for_object(self.entity.name, prop_name)
                value = config.label
            except ObjectDoesNotExist:
                value = prop_name
            cache.set(key, value)
            return value

    def __str__(self):
        return f"Object {self.uuid}"


class PropertyQuerySet(models.QuerySet):
    def for_object(self, obj):
        return self.filter(object_id=obj.uuid)

    def for_objects(self, objs):
        return self.filter(object_id__in=[obj.uuid for obj in objs])


class EntityConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(config_type=ElementConfig.ENTITY_CONFIG)

    @lru_cache()
    def for_entity(self, entity_name):
        return self.get_queryset().get(name=entity_name)


class RelationshipConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(config_type=ElementConfig.REL_CONFIG)

    @lru_cache()
    def get_by_name(self, relationship_name):
        return self.get_queryset().get(name=relationship_name)


class EntityPropertyConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.filter(config_type=ElementConfig.PROPERTY_CONFIG, parent_type=ElementConfig.ENTITY_CONFIG)

    @lru_cache()
    def for_object(self, parent_name, prop_name):
        return self.get_queryset().get(parent_name=parent_name, name=prop_name)


class RelPropertyConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.filter(config_type=ElementConfig.PROPERTY_CONFIG, parent_type=ElementConfig.REL_CONFIG)


class StatsConfigManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.filter(config_type=ElementConfig.STATS_CONFIG)

    def get_stats(self):
        return self.get(name="stats")


class ElementConfig(models.Model):
    ENTITY_CONFIG, REL_CONFIG, PROPERTY_CONFIG, STATS_CONFIG, CUSTOM_CONFIG = 1, 2, 3, 4, 5
    TYPE_CHOICES = [
        (ENTITY_CONFIG, "entity"),
        (REL_CONFIG, "relationship"),
        (PROPERTY_CONFIG, "property"),
        (STATS_CONFIG, "stats"),
        (CUSTOM_CONFIG, "custom"),
    ]
    PARENT_TYPE_CHOICES = [
        (ENTITY_CONFIG, "entity"),
        (REL_CONFIG, "relationship"),
    ]

    def __str__(self):
        text = f"Config: {self.get_config_type_display()}.{self.name}"
        if self.parent_type:
            text += f" ({self.get_parent_type_display()}.{self.parent_name})"
        return text

    config_type = models.PositiveSmallIntegerField(choices=TYPE_CHOICES)
    name = models.TextField()
    parent_type = models.PositiveSmallIntegerField(choices=PARENT_TYPE_CHOICES, null=True)
    parent_name = models.TextField(null=True)
    label = models.TextField()
    data = models.JSONField()

    class Meta:
        unique_together = ["config_type", "name", "parent_type", "parent_name"]
        verbose_name = "Element configuration"
        verbose_name_plural = "Element configurations"


class EntityConfig(ElementConfig):
    objects = EntityConfigManager()

    class Meta:
        proxy = True

    @property
    def label_properties(self):
        return self.data.get("label_properties", [])

    @property
    def label_template(self):
        return self.data.get("label_template", None)

    @property
    def graph_node_conf(self):
        return self.data.get("graph_node_conf", {})


class RelationshipConfig(ElementConfig):
    objects = RelationshipConfigManager()

    class Meta:
        proxy = True

    @property
    def from_graph_node_conf(self):
        return self.data.get("from_graph_node_conf", {})

    @property
    def to_graph_node_conf(self):
        return self.data.get("to_graph_node_conf", {})


class EntityPropertyConfig(ElementConfig):
    objects = EntityPropertyConfigManager()

    class Meta:
        proxy = True


class RelPropertyConfig(ElementConfig):
    objects = RelPropertyConfigManager()

    class Meta:
        proxy = True


class StatsConfig(ElementConfig):
    objects = StatsConfigManager()

    class Meta:
        proxy = True


class ObjectRepositoryQuerySet(models.QuerySet):
    def from_uuids(self, uuids):
        return self.filter(uuid__in=uuids).select_related("entity")

    def with_properties(self):
        # TODO: is this still needed?
        qs = []
        for obj in self.all():
            obj.__dict__["properties"] = obj.properties
            qs.append(obj)
        return qs

    def index(self, Model, start_id=None, lang="pg_catalog.portuguese", batch_size=10000, callback=None):
        """Update the FTS index using objects from `Model`"""

        table_name = Model._meta.db_table
        temp_table_name = "urlid_tmp_fts_" + random_name(16)
        create_temp_table_sqls = [
            (
                f'CREATE TEMPORARY TABLE "{temp_table_name}" ("id" SERIAL, "uuid" UUID)',
                [],
            ),
            (
                dedent(
                    f"""
                    INSERT INTO "{temp_table_name}" ("uuid")
                    SELECT DISTINCT object_uuid FROM "{table_name}" WHERE id >= %s
                    """
                ).strip(),
                [start_id or 0],
            ),
            (
                f'CREATE INDEX "idx_{random_name(20)}" ON "{temp_table_name}" ("id", "uuid")',
                [],
            ),
        ]
        row_count = []
        with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
            for sql, args in create_temp_table_sqls:
                cursor.execute(sql, args)
                row_count.append(cursor.rowcount)
        total_objects = row_count[1]

        string_agg = " || ' ' || ".join(
            f"COALESCE(STRING_AGG({field_name}, ' '), '')" for field_name in Model._meta.search_fields
        )
        insert_sql = dedent(
            f"""
            INSERT INTO "{ObjectRepository._meta.db_table}" (uuid, entity_id, search_data)
                SELECT
                    object_uuid,
                    %s::uuid,
                    to_tsvector(
                        %s,
                        {string_agg}
                    )
                FROM "{table_name}"
                WHERE object_uuid IN (
                    SELECT "uuid"
                    FROM "{temp_table_name}"
                    WHERE id >= %s AND id < %s
                )
                GROUP BY object_uuid
            ON CONFLICT (uuid) DO UPDATE SET search_data = EXCLUDED.search_data
            """
        ).strip()

        counter = 0
        for object_id in range(1, total_objects + 1, batch_size):
            args = (Model._meta.entity_uuid, lang, object_id, object_id + batch_size)
            with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
                cursor.execute(insert_sql, args)
                counter += cursor.rowcount
                if callback is not None:
                    callback(done=counter, total=total_objects)

        drop_temp_table_sql = f'DROP TABLE "{temp_table_name}"'
        with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
            cursor.execute(drop_temp_table_sql)

        return counter

    def search(self, search_query, config=urlid_graph_settings.SEARCH_LANGUAGE):
        """Full-Text Search in object's search_data"""

        qs = self.filter()
        search_query = search_query or ""
        if not search_query:
            return qs
        words = search_query.split()
        query = None
        done = []  # to preserve order
        for word in words:
            if not word or word in done:
                continue
            if query is None:
                query = SearchQuery(word, config=config)
            else:
                query = query & SearchQuery(word, config=config)
            done.append(word)
        qs = qs.annotate(search_rank=SearchRank(models.F("search_data"), query)).filter(search_data=query)
        # Using `qs.query.add_ordering` will APPEND ordering fields instead
        # of OVERWRITTING (`qs.order_by` will overwrite).
        qs.query.add_ordering("-search_rank")
        return qs

    def search_many_entities(self, query, entities=all, limit_per_entity=10):
        """Search objects in more than one entity, by name

        Prefer to use this method since it also filter objects by entity (more performance)"""

        if entities is all:
            entities = Entity.objects.all()
        else:
            entities = Entity.objects.by_name(entities)

        results = []
        for entity in entities:
            qs = self.filter(entity=entity).search(query).select_related("entity")
            results.extend(qs[:limit_per_entity])
        results.sort(key=lambda row: row.search_rank, reverse=True)
        return results


class ObjectRepository(ObjectMixin, models.Model):
    objects = ObjectRepositoryQuerySet.as_manager()

    uuid = models.UUIDField(primary_key=True)
    entity = models.ForeignKey(Entity, on_delete=models.DO_NOTHING, db_constraint=False, db_index=False)
    search_data = SearchVectorField(null=True)

    class Meta:
        indexes = [GinIndex(fields=["search_data"])]
        verbose_name = "Object repository"
        verbose_name_plural = "Object repositories"


class SavedGraph(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=255)
    edges = ArrayField(models.CharField(max_length=255))
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return f"{self.user.username} - {self.name}"

    class Meta:
        unique_together = (("user", "name"),)
        verbose_name = "Saved graph"
        verbose_name_plural = "Saved graphs"


class ObjectModelMixin(models.Model):
    object_uuid = models.UUIDField(null=False, blank=False, db_index=True)
    updated_at = models.DateTimeField(null=True, blank=True)

    def raw_serialize(self):
        return {field.name: getattr(self, field.name) for field in self._meta.fields}

    def serialize(self):
        return {field.verbose_name: getattr(self, field.name) for field in self._meta.fields}

    class Meta:
        abstract = True


class DatasetModel(models.Model):
    object_uuid = models.UUIDField(
        verbose_name="Object UUID",
        help_text="UUID of described object",
        null=False,
        blank=False,
        db_index=True,
    )

    @classmethod
    def extra(cls, qs):
        return {}

    @classmethod
    def fields(cls):
        from django.db import connection

        return [
            {
                "name": field.name,
                "title": field.verbose_name,
                "type": field.db_type(connection),
                "description": field.help_text,
            }
            for field in cls._meta.fields
            if field.name not in ("id", "object_uuid")
        ]

    @cached_property
    def serializers(self):
        return {
            attribute[len("serialize_") :]: getattr(self, attribute)
            for attribute in dir(self)
            if attribute.startswith("serialize_")
        }

    def serialize(self):
        row = []
        for field in type(self).fields():
            field_name = field["name"]
            value = getattr(self, field["name"])
            if field_name in self.serializers:
                value = self.serializers[field_name](value)
            row.append(value)
        return row

    @classmethod
    def subclasses(cls):
        data = {}
        for klass in cls.__subclasses__():
            if hasattr(klass._meta, "slug"):
                slug = klass._meta.slug
            else:
                slug = slug_from_model_name(klass.__name__)
            data[slug] = klass
        return data

    class Meta:
        abstract = True


class JobLog(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    description = models.CharField(max_length=255)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ("-created_at",)

    @property
    def current_step(self):
        return self.steps.first().step

    @property
    def progress(self):
        step = self.steps.first()
        if step.done is not None and step.total is not None and step.total != 0:
            return f"{100 * step.done / step.total:5.2f}%"
        else:
            result = step.step + (f" ({step.action})" if step.action else "")
            if self.steps.filter(action="error").count() > 0:
                result += " - has error"
            return result

    @property
    def eta(self):
        step = self.steps.first()
        if step.done is None or step.total is None or step.total == 0:
            return "-"
        first_step = self.steps.exclude(pk=step.pk).filter(step=step.step, done__gt=0).order_by("created_at").first()
        if first_step is None:
            return "-"
        to_finish = step.total - step.done
        delta_done = step.done - first_step.done
        delta_time = (step.created_at - first_step.created_at).total_seconds()
        speed = delta_done / delta_time
        result = str(datetime.timedelta(seconds=to_finish / speed))
        return result.split(".")[0]

    @property
    def last_updated_at(self):
        return self.steps.first().created_at

    def __str__(self):
        return f"{self.name} - {self.description}: {self.id}"

    def save(self, *args, **kwargs):
        if self._state.adding:
            super().save(*args, **kwargs)
            LogStep.objects.create(job=self, action="started", step="created")
        else:
            super().save(*args, **kwargs)


class LogStep(models.Model):
    job = models.ForeignKey(JobLog, related_name="steps", on_delete=models.CASCADE)
    action = models.CharField(max_length=255)
    step = models.CharField(max_length=255)
    done = models.FloatField(null=True, blank=True)
    total = models.FloatField(null=True, blank=True)
    message = models.TextField(null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ("-created_at",)

    def __str__(self):
        return f"{self.step} - {self.action}: {self.job.id}"



import traceback
from textwrap import dedent

from django.apps import apps
from django.core.management.base import BaseCommand
from django.db import connections
from urlid_graph import settings as urlid_graph_settings
from urlid_graph.log import IndexObjectsLogger
from urlid_graph.models import ObjectModelMixin, ObjectRepository
from urlid_graph.utils import random_name


class Command(BaseCommand):
    help = "Index objects from a specific model"

    def add_arguments(self, parser):
        parser.add_argument("--start-id", type=int)
        parser.add_argument("--batch-size", type=int, default=100_000)
        parser.add_argument("--no-delete", action="store_true")
        parser.add_argument("--only-missing", action="store_true")
        parser.add_argument("app_name")
        parser.add_argument("model")

    def handle(self, *args, **options):
        app_name = options["app_name"]
        batch_size = options["batch_size"]
        model = options["model"]
        only_missing = options["only_missing"]
        should_delete = not options["no_delete"] and not only_missing
        start_id = options["start_id"]

        Model = apps.get_model(app_name, model)
        if not issubclass(Model, ObjectModelMixin):
            raise ValueError("Model '{}' doesn't inherit from `ObjectModelMixin`".format(model))
        elif only_missing is True and start_id is not None:
            raise ValueError("--only-missing and --start-id are exclusive options")

        self.logger = IndexObjectsLogger(description=f"{app_name}.{model}")
        print(f"Starting job {self.logger.job.id}")

        ok = True
        try:
            if should_delete:
                self.logger.start_step("delete-objects")
                result = ObjectRepository.objects.filter(entity_id=Model._meta.entity_uuid).delete()
                self.logger.finish_step("delete-objects", message=f"{result[0]} objects deleted")

            self.logger.start_step("index")
            self.index(
                Model,
                only_missing=only_missing,
                start_id=start_id,
                batch_size=batch_size,
                callback=self.logger.progress,
            )
            self.logger.finish_step("index")

        except:  # noqa
            self.logger.error(traceback.format_exc())
            ok = False

        else:
            self.logger.finish()

        return str(ok)  # Used by import_data when calling this command programatically

    def index(
        self, Model, only_missing=False, start_id=None, lang="pg_catalog.portuguese", batch_size=10000, callback=None
    ):
        """Update the FTS index using objects from `Model`"""
        table_name = Model._meta.db_table
        temp_table_name = "urlid_tmp_fts_" + random_name(16)
        if only_missing:
            objrepo_table_name = ObjectRepository._meta.db_table
            select_uuids_sql = dedent(
                f"""
                SELECT
                  "m"."object_uuid" AS "uuid"
                FROM "{table_name}" AS "m"
                  LEFT JOIN "{objrepo_table_name}" AS "o"
                    ON "m"."object_uuid" = "o"."uuid"
                WHERE "o"."uuid" IS NULL;
            """
            ).strip()
            uuids_args = []
        else:
            select_uuids_sql = f"""
                SELECT DISTINCT object_uuid FROM "{table_name}" WHERE id >= %s
            """.strip()
            uuids_args = [start_id or 0]
        create_temp_table_sqls = [
            (
                f'CREATE TEMPORARY TABLE "{temp_table_name}" ("id" SERIAL, "uuid" UUID)',
                [],
            ),
            (
                dedent(
                    f"""
                    INSERT INTO "{temp_table_name}" ("uuid")
                    {select_uuids_sql}
                    """
                ).strip(),
                uuids_args,
            ),
            (
                f'CREATE INDEX "idx_{random_name(20)}" ON "{temp_table_name}" ("id", "uuid")',
                [],
            ),
        ]
        row_count = []
        with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
            for sql, args in create_temp_table_sqls:
                cursor.execute(sql, args)
                row_count.append(cursor.rowcount)
        total_objects = row_count[1]

        string_agg = " || ' ' || ".join(
            f"COALESCE(STRING_AGG({field_name}, ' '), '')" for field_name in Model._meta.search_fields
        )
        insert_sql = dedent(
            f"""
            INSERT INTO "{ObjectRepository._meta.db_table}" (uuid, entity_id, search_data)
                SELECT
                    object_uuid,
                    %s::uuid,
                    to_tsvector(
                        %s,
                        {string_agg}
                    )
                FROM "{table_name}"
                WHERE object_uuid IN (
                    SELECT "uuid"
                    FROM "{temp_table_name}"
                    WHERE id >= %s AND id < %s
                )
                GROUP BY object_uuid
            ON CONFLICT (uuid) DO UPDATE SET search_data = EXCLUDED.search_data
            """
        ).strip()

        counter = 0
        for object_id in range(1, total_objects + 1, batch_size):
            args = (Model._meta.entity_uuid, lang, object_id, object_id + batch_size)
            with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
                cursor.execute(insert_sql, args)
                counter += cursor.rowcount
                if callback is not None:
                    callback(done=counter, total=total_objects)

        drop_temp_table_sql = f'DROP TABLE "{temp_table_name}"'
        with connections[urlid_graph_settings.DJANGO_DATABASE].cursor() as cursor:
            cursor.execute(drop_temp_table_sql)

        return counter



import pprint

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connections
from django.utils import timezone

from urlid_graph.log import AsyncLogger, Step
from urlid_graph.models import ElementConfig, Entity


class DataStats:
    def __init__(self):
        graph_django_connection = connections[settings.GRAPH_DATABASE]
        graph_django_connection.connect()
        self.graph_connection = graph_django_connection.connection

        self._total_objects_per_entity = None
        self._total_relationships_per_type = None

    @property
    def total_objects(self):
        return sum(self.total_objects_per_entity.values())

    @property
    def total_objects_per_entity(self):
        if self._total_objects_per_entity is None:
            self._total_objects_per_entity = {}
            for entity in Entity.objects.all():
                Model = entity.get_model()
                if Model is None:
                    continue
                self._total_objects_per_entity[entity.name] = Model.objects.values("object_uuid").distinct().count()

        return self._total_objects_per_entity

    @property
    def total_relationships(self):
        return sum(self.total_relationships_per_type.values())

    @property
    def total_relationships_per_type(self):
        if self._total_relationships_per_type is None:
            with self.graph_connection.cursor() as cursor:
                cursor.execute("MATCH ()-[r]->() RETURN type(r), COUNT(type(r))")
                data = dict(cursor.fetchall())
                self._total_relationships_per_type = {
                    key.replace('"', "").replace("'", ""): int(value) if value else None for key, value in data.items()
                }
        return self._total_relationships_per_type


class CollectStatsLogger(AsyncLogger):
    name = "update stats"
    steps = [
        Step("object-total", "Count total objects"),
        Step("object-entities", "Count total objects per entity"),
        Step("relationship-total", "Count total relationships"),
        Step("relationship-entities", "Count total relationships per entity"),
    ]


class Command(BaseCommand):
    help = "Update general statistics about imported data"

    def handle(self, *args, **options):
        self.logger = CollectStatsLogger(description="collecting statistics from objects and relationships")
        print(f"Starting job {self.logger.job.id}")

        data = {
            "object": {
                "total": None,
                "entities": {},
            },
            "relationship": {
                "total": None,
                "entities": {},
            },
        }
        stats = DataStats()

        self.logger.start_step("object-total")
        data["object"]["total"] = stats.total_objects
        self.logger.finish_step("object-total", message=f"Found {data['object']['total']}")

        self.logger.start_step("object-entities")
        data["object"]["entities"].update(stats.total_objects_per_entity)
        entity_total = [f"{key}: {value}" for key, value in data["object"]["entities"].items()]
        self.logger.finish_step("object-entities", message=f"Found: {', '.join(entity_total)}")

        self.logger.start_step("relationship-total")
        data["relationship"]["total"] = stats.total_relationships
        self.logger.finish_step("relationship-total", message=f"Found {data['relationship']['total']}")

        self.logger.start_step("relationship-entities")
        data["relationship"]["entities"].update(stats.total_relationships_per_type)
        relationship_total = [f"{key}: {value}" for key, value in data["relationship"]["entities"].items()]
        self.logger.finish_step("relationship-entities", message=f"Found: {', '.join(relationship_total)}")

        data["updated_at"] = str(timezone.now())
        pprint.pprint(data)

        ElementConfig.objects.update_or_create(
            config_type=ElementConfig.STATS_CONFIG,
            name="stats",
            parent_type=None,
            parent_name=None,
            defaults={"label": "stats", "data": data},
        )
        self.logger.finish()
